#!/usr/bin/env python3
import abc
from dataclasses import dataclass
from typing import Optional, Union

import torch.nn as nn
from tensordict import TensorDict
from torchrl.data import BoundedTensorSpec, CompositeSpec, DiscreteTensorSpec
from utils import ReplayBuffer


EvalMode = bool
T0 = bool


@dataclass
class AgentConfig:
    """Base config for agents"""

    _target_: str
    device: str = "cuda"
    name: str = "BaseAgent"


class Agent(abc.ABC, nn.Module):
    def __init__(
        self,
        cfg: AgentConfig,
        obs_spec: CompositeSpec,
        act_spec: Union[DiscreteTensorSpec, BoundedTensorSpec],
    ):
        super().__init__()

        self.cfg = cfg

        self.obs_spec = obs_spec
        self.act_spec = act_spec

        if isinstance(act_spec, DiscreteTensorSpec):
            self.discrete_acts = True
            raise NotImplementedError("Need to implement discrete action space")
        else:
            self.discrete_acts = False
            self.register_buffer("act_spec_low", act_spec.low.to(cfg.device))
            self.register_buffer("act_spec_high", act_spec.high.to(cfg.device))

    @abc.abstractmethod
    def select_action(
        self, obs: TensorDict, t0: bool = False, eval_mode: EvalMode = False
    ):
        raise NotImplementedError

    @abc.abstractmethod
    def update(
        self, replay_buffer: ReplayBuffer, num_new_transitions: int
    ) -> Optional[dict]:
        raise NotImplementedError

    @property
    def device(self):
        return self.cfg.device

    @property
    def name(self):
        return self.cfg.name

    @property
    def total_params(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)
