#!/usr/bin/env python3
import logging

import helper as h
import torch
import torch.nn as nn
from omegaconf import DictConfig


logger = logging.getLogger(__name__)


class Actor(nn.Module):
    def __init__(
        self,
        cfg: DictConfig,
        # cfg: FSQMPCConfig,
        state_dim: int,
        act_dim: int,
        action_scale,
        action_bias,
        act_low,
        act_high,
    ):
        super().__init__()
        self.cfg = cfg
        self.action_scale = action_scale
        self.action_bias = action_bias
        self.act_low = act_low
        self.act_high = act_high
        self.mlp = h.mlp(state_dim, self.cfg.mlp_dims, act_dim)

    def forward(self, z):
        a = self.mlp(z)
        a = torch.tanh(a)
        a = a * self.action_scale + self.action_bias
        return a


class Critic(nn.Module):
    def __init__(
        self,
        cfg: DictConfig,
        # cfg: FSQMPCConfig,
        state_dim: int,
        act_dim: int,
    ):
        super().__init__()
        self.cfg = cfg
        qs = [
            h.mlp(state_dim + act_dim, mlp_dims=cfg.mlp_dims, out_dim=1).to(cfg.device)
            for _ in range(cfg.num_critics)
        ]
        for q in qs:
            h.orthogonal_init(q.parameters())

        self.qs = h.Ensemble(qs)

    def forward(self, z, a, return_type: str = "all"):
        x = torch.cat([z, a], -1)
        qs = self.qs(x)
        if return_type == "all":
            return qs

        # Sample two Q values
        if self.cfg.q_sample_size is not None:
            idxs = torch.randperm(qs.shape[0])[: self.cfg.q_sample_size]
            qs = qs[idxs]
        if return_type == "min":
            return torch.min(qs, 0)[0]
        elif return_type == "avg":
            return torch.mean(qs, 0)
        else:
            raise NotImplementedError(
                f"return_type should be 'all' or 'min' or 'avg' not {return_type}"
            )
