#!/usr/bin/env python3
import copy
import logging
import math
from dataclasses import dataclass, field
from typing import List, Optional, Tuple

import helper as h
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import wandb
from helper import soft_update_params
from tensordict import TensorDict
from torchrl.data import BoundedTensorSpec, CompositeSpec
from tqdm import trange
from utils import ReplayBuffer, ReplayBufferSamples

from .base import Agent, AgentConfig, EvalMode
from .sac import SAC
from .td3 import TD3


logger = logging.getLogger(__name__)


@dataclass
class FSQMPCConfig(AgentConfig):
    """Config for DCWM"""

    """RL Algorithm"""
    rl: str = "TD3"  # Either "TD3" or "SAC"

    """WORLD MODEL CONFIG"""
    """Size of latent space"""
    latent_dim: int = 512
    """Horizon used for representation learning"""
    horizon: int = 5
    """Discount factor for representation learning"""
    rho: float = 0.9
    """MLP dims for encoder/dynamics/reward/decoder"""
    enc_mlp_dims: List[int] = field(default_factory=lambda: [256])
    """Learning rate for encoder/dynamics/projection/reward"""
    enc_lr: float = 1e-4
    """Momentum coefficient for target encoder"""
    enc_tau: float = 0.005
    """Update encoder less frequently than actor/critic"""
    enc_update_freq: int = 1
    """Clips the gradient norm of the encoder"""
    grad_clip_norm: Optional[int] = 20
    """Use target encoder for representation learning"""
    use_tar_enc: bool = True
    """Predict change in latent or next latent? i.e. next_z = z + f(z, a) else next_z = f(z, a)"""
    use_delta: bool = True
    """Use LayerNorm or BatchNorm for encoder?"""
    enc_norm_type: str = "ln"
    """Use updated encoder for actor/critic training?"""
    use_updated_encoder_for_actor_critic: bool = True
    """Use temporal consistency loss for representation learning"""
    use_tc_loss: bool = True
    """Use reward prediction for representation learning"""
    use_rew_loss: bool = False
    """Use cosine similarity for consistency loss - otherwise MSE"""
    use_cosine_similarity_dynamics: bool = True
    reward_coef: float = 1.0
    consistency_coef: float = 1.0
    r_min: Optional[float] = None  # min reward, e.g. 0 in DMC
    r_max: Optional[float] = None  # max reward, e.g. 1 in DMC
    discrete_encoding_type: str = "codes"  # "codes" or "indices" or "one-hot"
    """Flag to turn FSQ on/off """
    use_fsq: bool = True
    """FSQ levels - setting as [8,8] corresponds to a codebook of size 8*8=62=2^8"""
    fsq_levels: List[int] = field(default_factory=lambda: [8, 8])
    """PROJECTION HEAD"""
    """Project the latent using an MLP before calculating the temporal consistency loss?"""
    use_latent_projection: bool = False
    """MLP dims for projection head"""
    projection_mlp_dims: List[int] = field(default_factory=lambda: [256])
    """Dimension of projection - defaults to latent_dim/16"""
    proj_dim: Optional[int] = None

    use_dynamics_sample_for_actor_critic: bool = False

    """Planning"""
    mpc: bool = False
    iterations: int = 6
    num_samples: int = 512
    num_elites: int = 64
    num_pi_trajs: int = 24
    plan_horizon: int = 3
    min_std: float = 0.05
    max_std: int = 2
    temperature: float = 0.5
    plan_ce_selection: Optional[str] = None
    """"(Optionally) Add TD3 exploration noise to MPPI action or use instead of MPPI std"""
    add_exploration_noise_after_mppi: Optional[str] = None  # None or "add" or "instead"
    add_exploration_noise_inside_mppi: Optional[
        str
    ] = None  # None or "instead" or "max"
    mppi_use_mean: bool = False
    use_prior_policy_noise: bool = True

    """Dynamics"""
    ce_selection: str = "sample"
    use_ce_loss_dynamics: bool = False
    use_ce_loss_dynamics_softmax: bool = False

    """OTHER"""
    """Logging frequency"""
    logging_freq: int = 100
    """If True try to compile all NNs"""
    compile: bool = False
    """All NNs will be put on this device"""
    device: str = "cuda"
    name: str = "FSQMPC"


class WorldModel(nn.Module):
    def __init__(
        self, cfg: FSQMPCConfig, obs_spec: CompositeSpec, act_spec: BoundedTensorSpec
    ):
        super().__init__()
        self.cfg = cfg
        self.obs_spec = obs_spec
        self.act_spec = act_spec
        obs_dim = np.array(obs_spec["state"].shape).prod().item()
        act_dim = np.array(act_spec.shape).prod().item()

        ##### Configure FSQ stuff #####
        self.org_latent_dim = copy.copy(cfg.latent_dim)
        self.enc_latent_dim = copy.copy(cfg.latent_dim)
        if cfg.use_fsq:
            self.num_channels = len(cfg.fsq_levels)
            if not cfg.latent_dim % self.num_channels == 0:
                raise NotImplementedError(
                    "latent_dim must be divisible by number of FSQ channels"
                )
            self._fsq = h.FSQ(levels=cfg.fsq_levels)

            self.enc_latent_dim = self.org_latent_dim * self.num_channels

            if self.cfg.discrete_encoding_type == "codes":
                self.cfg.latent_dim *= self.num_channels
            elif self.cfg.discrete_encoding_type == "indices":
                assert cfg.use_ce_loss_dynamics
                pass
            elif self.cfg.discrete_encoding_type == "one-hot":
                assert cfg.use_ce_loss_dynamics
                self.cfg.latent_dim *= self._fsq._fsq.codebook_size
            else:
                raise NotImplementedError(
                    f"discrete_encoding_type should be either 'codes', 'indices' or 'one-hot', not {self.cfg.discrete_encoding_type}"
                )

        ##### Init encoder #####
        self._encoder = nn.ModuleDict()
        if "state" in cfg.obs_types:  # Encoder for state-based observations
            self._encoder.update(
                {"state": h.mlp(obs_dim, cfg.enc_mlp_dims, self.enc_latent_dim)}
            )
            if cfg.compile:
                self._encoder["state"] = torch.compile(
                    self._encoder["state"], mode="default"
                )
        if "pixels" in cfg.obs_types:  # Encoder for pixel-based observations
            if self.cfg.enc_norm_type == "bn":
                raise NotImplementedError("Need to implement BN for CNN encoder")
            self._encoder.update(
                {
                    "pixels": h.CNNEncoder(
                        obs_shape=obs_spec.shape,
                        latent_dim=self.enc_latent_dim,
                        hidden_dim=256,
                        frame_diff=False,
                    )
                }
            )
        if cfg.use_tar_enc:
            self._encoder_tar = copy.deepcopy(self._encoder).requires_grad_(False)

        # If training dynamics w/ cross entropy change output dim
        trans_out_dim = self.cfg.latent_dim
        if self.cfg.use_ce_loss_dynamics and self.cfg.use_ce_loss_dynamics_softmax:
            assert cfg.use_fsq
            trans_out_dim = int(self.org_latent_dim * self._fsq._fsq.codebook_size)
        self._trans = h.mlp(self.enc_latent_dim + act_dim, cfg.mlp_dims, trans_out_dim)
        if cfg.compile:
            self._trans = torch.compile(self._trans, mode="default")

        if cfg.use_latent_projection:
            if cfg.proj_dim is None:
                cfg.proj_dim = int(self.cfg.latent_dim / 16)
            self._proj = h.mlp(cfg.latent_dim, cfg.mlp_dims, cfg.proj_dim)
            if cfg.use_tar_enc:
                self._proj_tar = copy.deepcopy(self._proj).requires_grad_(False)

        if cfg.use_rew_loss:
            self._reward = h.mlp(self.cfg.latent_dim + act_dim, cfg.mlp_dims, 1)
            if cfg.compile:
                self._reward = torch.compile(self._reward, mode="default")

            if cfg.r_max is not None and cfg.r_min is not None:
                r_scale = (cfg.r_max - cfg.r_min) / 2.0
                r_bias = (cfg.r_max + cfg.r_min) / 2.0
                self.r_scale_fn = lambda r: torch.tanh(r) * r_scale + r_bias
            else:
                self.r_scale_fn = lambda r: r

    def encode(self, obs, tar: bool = False):
        if "pixels" in self.cfg.obs_types:
            raise NotImplementedError()
        zs = {}
        for key in obs.keys():
            if tar:
                zs.update({key: self._encoder_tar[key](obs[key])})
            else:
                zs.update({key: self._encoder[key](obs[key])})
        if "state" in self.cfg.obs_types and "pixels" not in self.cfg.obs_types:
            z = zs["state"]
            td = TensorDict({"state": z}, batch_size=obs.batch_size)
        elif "state" not in self.cfg.obs_types and "pixels" in self.cfg.obs_types:
            z = zs["pixels"]
        else:
            raise NotImplementedError("Need to make encoder take both state and pixels")

        td = TensorDict({"state": z}, batch_size=obs.batch_size)
        if self.cfg.use_fsq:
            td.update(self.quantize(z))
        else:
            td.update({"codes": z})
        return td

    def trans(self, z, a, ce_selection: Optional[str] = None):
        za = torch.concat([z, a], -1)

        if self.cfg.use_ce_loss_dynamics and self.cfg.use_ce_loss_dynamics_softmax:
            # Returns logits for each class
            logits = self._trans(za)
            logits = logits.reshape(
                -1,
                self.org_latent_dim,
                self._fsq._fsq.codebook_size,
            )

            if ce_selection is None:
                ce_selection = self.cfg.ce_selection

            # Convert latent state logits to an actual latent state
            if ce_selection in ["mode", "max"]:
                # TODO this is the same as the mode
                # Note this has no gradients so should only be used for MPC
                indices = torch.max(logits, -1)[1]
                next_z = self._fsq._fsq.implicit_codebook[
                    indices.to(torch.long)
                ].flatten(-2)
                next_z_dict = {
                    "codes": next_z,
                    "logits": logits,
                    "indices": indices,
                    # "one-hot": z_one_hot.flatten(-2),
                }
            elif "sample-no-grad" in ce_selection:
                z_one_hot = torch.nn.functional.gumbel_softmax(
                    logits, tau=1, hard=True, dim=-1
                )
                indices = torch.max(z_one_hot, -1)[1].to(torch.float)
                next_z = self._fsq._fsq.implicit_codebook[
                    indices.to(torch.long)
                ].flatten(-2)
                next_z_dict = {
                    "codes": next_z,
                    "logits": logits,
                    "indices": indices,
                    "one-hot": z_one_hot.flatten(-2),
                }
            elif "sample" in ce_selection:
                z_one_hot = torch.nn.functional.gumbel_softmax(
                    logits, tau=1, hard=True, dim=-1
                )
                codebook = self._fsq._fsq.implicit_codebook[None, None, ...]
                next_z = (z_one_hot[..., None, :] @ codebook)[..., 0, :].flatten(-2)
                indices = torch.max(z_one_hot, -1)[1].to(torch.float)
                next_z_dict = {
                    "codes": next_z,
                    "logits": logits,
                    "indices": indices,
                    "one-hot": z_one_hot.flatten(-2),
                }
            # elif "mode" in ce_selection:
            #     dist = torch.distributions.Categorical(logits=logits)
            #     # probs = dist.probs[..., None]
            #     # next_z
            #     indices = dist.mode
            elif "weighted-avg" in ce_selection:
                probs = F.softmax(logits, dim=-1)[..., None]
                codebook = self._fsq._fsq.implicit_codebook[None, None, ...]
                next_z = (probs * codebook).sum(-2)
                next_z = next_z.reshape(-1, self.org_latent_dim * self.num_channels)
                next_z_dict = {"codes": next_z, "logits": logits}
            else:
                raise NotImplementedError
            # if "weighted-avg" not in self.cfg.ce_selection:
            #     # No gradient through this...
            #     next_z = self._fsq.indices_to_codes(indices).view(
            #         -1, self.cfg.latent_dim
            #     )
            if self.cfg.discrete_encoding_type == "one-hot":
                if "one-hot" not in next_z_dict.keys():
                    next_z_dict.update(
                        {
                            "one-hot": torch.nn.functional.gumbel_softmax(
                                logits, tau=1, hard=True, dim=-1
                            ).flatten(-2)
                        }
                    )
            if self.cfg.discrete_encoding_type == "indices":
                if "indices" not in next_z_dict.keys():
                    next_z_dict.update({"indices": torch.max(logits, -1)[1].float()})
        else:
            delta_z = self._trans(za)
            next_z = z + delta_z if self.cfg.use_delta else delta_z
            if self.cfg.use_fsq:
                next_z = self.quantize(next_z)["codes"]

            next_z_dict = {"codes": next_z}

        next_z_dict.update(
            {
                "z": next_z.reshape(
                    *next_z.shape[0:-1], self.org_latent_dim, self.num_channels
                )
            }
        )
        return TensorDict(
            next_z_dict,
            batch_size=torch.Size([z.shape[0]]),
            device=self.cfg.device,
        )

    def reward(self, z, a):
        za = torch.concat([z, a], -1)
        r = self._reward(za)
        r = self.r_scale_fn(r)
        return r

    def project(self, z, tar: bool = False):
        """Project latent state before calculating consistency loss"""
        if tar and self.cfg.use_tar_enc:
            z = self._proj_tar(z)
        else:
            z = self._proj(z)
        return z

    def quantize(self, z):
        """Quantize the latent state"""
        # return self._fsq(z)
        td = self._fsq(z)
        if self.cfg.discrete_encoding_type == "codes":
            td["state"] = td["codes"]
        elif self.cfg.discrete_encoding_type == "indices":
            td["state"] = td["indices"].to(torch.float)
        elif self.cfg.discrete_encoding_type == "one-hot":
            td["state"] = (
                F.one_hot(
                    td["indices"].to(torch.long),
                    num_classes=self._fsq._fsq.codebook_size,
                )
                .flatten(-2)
                .to(torch.float)
            )
        else:
            raise NotImplementedError(
                f"discrete_encoding_type should be either 'codes', 'indices' or 'one-hot', not {self.cfg.discrete_encoding_type}"
            )
        return td

    def loss(self, batch: ReplayBufferSamples) -> Tuple[torch.Tensor, dict]:
        tc_loss = torch.zeros(1).to(self.cfg.device)
        reward_loss = torch.zeros(1).to(self.cfg.device)

        a = batch.actions

        ##### Create targets #####
        with torch.no_grad():
            next_obs = batch.next_observations
            zs_tar = self.encode(next_obs, tar=True)

        ##### Latent rollout #####
        zs = {
            "codes": torch.empty(
                self.cfg.horizon + 1,
                self.cfg.batch_size,
                # self.cfg.latent_dim,
                self.enc_latent_dim,
                device=self.cfg.device,
            )
        }
        if self.cfg.use_ce_loss_dynamics:
            zs.update(
                {
                    "logits": torch.empty(
                        self.cfg.horizon + 1,
                        self.cfg.batch_size,
                        self.org_latent_dim,
                        # int(self.cfg.latent_dim / self.num_channels),
                        self._fsq._fsq.codebook_size,
                        device=self.cfg.device,
                    )
                }
            )
        zs = TensorDict(
            zs,
            batch_size=torch.Size([self.cfg.horizon + 1, self.cfg.batch_size]),
            device=self.cfg.device,
        )
        z = self.encode(batch.observations[0])["codes"]
        zs["codes"][0] = z
        # zs[0] = z
        dones = torch.zeros_like(batch.dones[0], dtype=torch.bool)
        terminateds_or_dones = torch.zeros_like(batch.dones, dtype=torch.bool)

        for t in range(self.cfg.horizon):
            dones = torch.where(terminateds_or_dones[t], dones, batch.dones[t])
            terminateds_or_dones[t] = torch.logical_or(
                terminateds_or_dones[t], torch.logical_or(dones, batch.terminateds[t])
            )

            # Predict next latent
            next_z = self.trans(z=z, a=a[t])
            zs[t + 1] = next_z

            # Don't forget this
            z = next_z["codes"]

        rho = torch.tensor([self.cfg.rho**t for t in range(self.cfg.horizon)]).to(
            self.cfg.device
        )
        dones = batch.dones.to(torch.int)
        # terminateds_or_dones = terminateds_or_dones.to(torch.int)

        ##### (Optional) Reward prediction loss #####
        if self.cfg.use_rew_loss:
            r_tar = batch.rewards[..., None]  # Reward target
            r_pred = self.reward(z=zs[self.cfg.discrete_encoding_type][:-1], a=a)
            assert r_pred.ndim == 3 and r_tar.ndim == 3
            _reward_loss = (r_pred[..., 0] - r_tar[..., 0]) ** 2
            _rho_reward_loss = rho * torch.mean(
                # (1 - terminateds_or_dones) * _reward_loss, -1
                (1 - dones) * _reward_loss,
                -1,
            )
            reward_loss = torch.mean(_rho_reward_loss)

        ##### (Optional) Project latent before consistency loss #####
        if self.cfg.use_latent_projection:
            zs_tar["codes"] = self.project(zs_tar["codes"], tar=self.cfg.use_tar_enc)
            zs["codes"] = self.project(zs["codes"], tar=False)

        ##### Temporal consistency loss #####
        if self.cfg.use_tc_loss:
            if self.cfg.use_ce_loss_dynamics:
                """Cross entropy"""
                if not self.cfg.use_ce_loss_dynamics_softmax:
                    """If not using classification for dynamics use alternative to get logits"""
                    zs_ = zs["codes"][1:].view(
                        self.cfg.horizon,
                        self.cfg.batch_size,
                        int(self.cfg.latent_dim / self.num_channels),
                        self.num_channels,
                    )[..., None, :]
                    codebook = self._fsq._fsq.implicit_codebook[None, None, None, ...]
                    if self.cfg.use_cosine_similarity_dynamics:
                        """Cosine similarity with codebook"""
                        # TODO use compute_logits like CLIP
                        # org_latent_dim = int(self.cfg.latent_dim / self.num_channels)
                        # zs_logits = self.compute_logits(
                        #     zs_,
                        #     torch.broadcast_to(
                        #         codebook,
                        #         (zs_.shape[0], org_latent_dim, -1, -1),
                        #     ),
                        # )[..., 0, :]
                        zs["logits"][1:] = nn.CosineSimilarity(dim=-1, eps=1e-6)(
                            zs_, codebook
                        )
                    else:
                        """Inner product with codebook"""
                        zs["logits"][1:] = torch.einsum(
                            "hbdic,hbdCc->hbdC", zs_, codebook
                        )

                _tc_loss = torch.vmap(torch.vmap(F.cross_entropy))(
                    zs["logits"][1:], zs_tar["indices"].to(torch.long)
                )
            else:
                if self.cfg.use_cosine_similarity_dynamics:
                    """Cosine similarity"""
                    _tc_loss = -nn.CosineSimilarity(dim=-1, eps=1e-6)(
                        zs["codes"][1:], zs_tar["codes"]
                    )

                else:
                    """Mean squared error"""
                    _tc_loss = torch.mean(
                        (zs["codes"][1:] - zs_tar["codes"]) ** 2, dim=-1
                    )

            _rho_tc_loss = rho * torch.mean((1 - dones) * _tc_loss, -1)
            tc_loss = torch.mean(_rho_tc_loss)

        # if batch.terminateds[0:2, :].max() == 1:
        #     breakpoint()
        loss = self.cfg.consistency_coef * tc_loss + self.cfg.reward_coef * reward_loss
        info = {
            "tc_loss": tc_loss.item(),
            "reward_loss": reward_loss.item(),
            "enc_loss": loss.item(),
            "z_min": torch.min(zs["codes"]).item(),
            "z_max": torch.max(zs["codes"]).item(),
            "z_mean": torch.mean(zs["codes"].to(torch.float)).item(),
            "z_median": torch.median(zs["codes"]).item(),
        }
        if self.cfg.use_rew_loss:
            info.update(
                {
                    "r_min": r_pred.min().item(),
                    "r_max": r_pred.max().item(),
                    "r_mean": r_pred.mean().item(),
                }
            )
        return loss, info

    def soft_update_params(self):
        # Update the tar network
        if self.cfg.use_tar_enc:
            soft_update_params(self._encoder, self._encoder_tar, tau=self.cfg.enc_tau)
            if self.cfg.use_latent_projection:
                soft_update_params(self._proj, self._proj_tar, tau=self.cfg.enc_tau)

    def metrics(self, batch):
        z = self.encode(batch.observations[0])

        # Calculate rank of latent
        metrics = h.calc_rank(name="z", z=z["state"])

        # Calculate percentage of codebook that's active
        if self.cfg.use_fsq:
            num_codes = torch.tensor(math.prod(self.cfg.fsq_levels), device=z.device)
            # active_percent = z["indices"].unique().numel() / num_codes * 100

            def act_percent_fn(z):
                # TODO can't vmap this because Tensor.unique() not allowed in vmap
                return z.unique().numel() / num_codes * 100

            active_percents = torch.empty(z["indices"].shape[1])
            for i in range(z["indices"].shape[1]):
                active_percents[i] = act_percent_fn(z["indices"][i])
            metrics.update(
                {
                    # "active_percent": active_percent,
                    "active_percent_avg": active_percents.mean(),
                    "active_percent_min": active_percents.min(),
                    "active_percent_max": active_percents.max(),
                }
            )

        # TODO add dormant neuron ratio stuff
        # metrics.update(h.calc_dormant_neuron_ratio(batch, agent=self))

        return metrics

    @property
    def total_params(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)


class FSQMPC(Agent):
    def __init__(
        self, cfg: FSQMPCConfig, obs_spec: CompositeSpec, act_spec: BoundedTensorSpec
    ):
        super().__init__(cfg=cfg, obs_spec=obs_spec, act_spec=act_spec)
        self.cfg = cfg
        self.model = WorldModel(cfg, obs_spec=obs_spec, act_spec=act_spec).to(
            cfg.device
        )
        if cfg.compile:
            self.model = torch.compile(self.model, mode="default")
        self.model_opt = torch.optim.AdamW(self.model.parameters(), lr=cfg.enc_lr)

        if "pixels" in cfg.obs_types:
            raise NotImplementedError

        ##### Calculate dimensions for MLPs #####
        self.act_dim = np.array(act_spec.shape).prod().item()
        if "state" in cfg.obs_types:
            obs_dim = np.array(obs_spec["state"].shape).prod().item()
        else:
            raise NotImplementedError("Need to use state observations")

        ##### Init base RL algorithm #####
        if "TD3" in cfg.rl:
            make_agent_fn = TD3
        elif "SAC" in cfg.rl:
            make_agent_fn = SAC
        else:
            raise NotImplementedError
        self.agent = make_agent_fn(
            cfg,
            obs_spec=obs_spec,
            act_spec=act_spec,
            state_dim=self.model.cfg.latent_dim,
        ).to(cfg.device)

    def update(self, replay_buffer: ReplayBuffer, num_new_transitions: int) -> dict:
        """Update representation and TD3 at same time"""
        n = int(num_new_transitions * self.cfg.utd_ratio)
        info = {}

        if self.cfg.use_dynamics_sample_for_actor_critic:

            def map_obs_to_latent(batch):
                # Map observations to latent
                with torch.no_grad():
                    z = self.model.encode(batch.observations, tar=False)
                    next_z = torch.vmap(self.model.trans, randomness="different")(
                        z["codes"], batch.actions, ce_selection="sample-no-grad"
                    )
                    # next_z["state"] = next_z["codes"]
                    next_z["state"] = next_z[self.cfg.discrete_encoding_type]
                    del next_z["logits"]
                batch = batch._replace(observations=z, next_observations=next_z)
                return batch

        else:

            def map_obs_to_latent(batch):
                # Map observations to latent
                with torch.no_grad():
                    z = self.model.encode(batch.observations, tar=False)
                    next_z = self.model.encode(batch.next_observations, tar=False)
                batch = batch._replace(observations=z, next_observations=next_z)
                return batch

        for i in trange(n, desc="Training DCWM:", miniters=self.cfg.logging_freq):
            batch = replay_buffer.sample()

            if not self.cfg.use_updated_encoder_for_actor_critic:
                actor_critic_batch = map_obs_to_latent(batch)

            # Update enc less frequently than actor/critic
            if i % self.cfg.enc_update_freq == 0:
                info.update(self.model_update_step(batch=batch))

            if self.cfg.use_updated_encoder_for_actor_critic:
                actor_critic_batch = map_obs_to_latent(batch)

            ##### Make nstep returns #####
            # if self.cfg.horizon == 1 and self.cfg.nstep > 1:
            #     raise NotImplementedError("Check N-step batch is made correctly if h=1")

            ##### Update RL agent #####
            info.update(self.agent.update_step(batch=actor_critic_batch))

            if i % self.cfg.logging_freq == 0:
                if wandb.run is not None:
                    wandb.log(info)

        self.agent.post_update()
        return info

    def model_update_step(self, batch: ReplayBufferSamples):
        self.model.train()
        loss, info = self.model.loss(batch=batch)

        self.model_opt.zero_grad(set_to_none=True)
        loss.backward()

        if self.cfg.grad_clip_norm is not None:
            enc_params = list(self.model.parameters())
            grad_norm = torch.nn.utils.clip_grad_norm_(
                enc_params, self.cfg.grad_clip_norm, error_if_nonfinite=False
            )
            info.update({"grad_norm": float(grad_norm)})

        self.model_opt.step()

        # Update the tar network
        self.model.soft_update_params()

        if hasattr(self, "mppi_std"):
            info.update({f"mppi_std": self.mppi_std[0].mean().item()})
        self.model.eval()
        return info

    @torch.no_grad()
    def select_action(self, obs, t0: bool = False, eval_mode: EvalMode = False):
        is_flat_obs = False
        if obs.batch_size == torch.Size([]):
            obs = obs.view(1)
            is_flat_obs = True

        z = self.model.encode(obs, tar=False).to(torch.float)
        if self.cfg.mpc:
            a, self.mppi_std = self.plan(z, t0=t0, eval_mode=eval_mode)
        else:
            a = self.agent.pi(
                z[self.cfg.discrete_encoding_type], tar=False, eval_mode=eval_mode
            )
            a = a[0] if is_flat_obs else a

        return a

    @torch.no_grad()
    def plan(self, z, t0: bool = False, eval_mode=False):
        """
        Plan a sequence of actions using the learned world model.

        Args:
            z (torch.Tensor): Latent state from which to plan.
            t0 (bool): Whether this is the first observation in the episode.
            eval_mode (bool): Whether to use the mean of the action distribution.

        Returns:
            torch.Tensor: Action to take in the environment.
        """
        z_td = z
        z = z["state"]
        batch_size = z.shape[0]
        pi_actions = torch.empty(
            batch_size,
            self.cfg.plan_horizon,
            self.cfg.num_pi_trajs,
            self.act_dim,
            device=self.device,
        )
        actions = torch.empty(
            batch_size,
            self.cfg.plan_horizon,
            self.cfg.num_samples,
            self.act_dim,
            device=self.device,
        )
        mean = torch.zeros(
            batch_size, self.cfg.plan_horizon, self.act_dim, device=self.device
        )
        exploration_std = torch.tensor(
            [self.agent.exploration_noise] * self.cfg.plan_horizon,
            device=z.device,
            dtype=z.dtype,
        )[..., None]

        def single_mppi(z, actions, pi_actions, mean, prev_mean):
            # Sample policy trajectories
            if self.cfg.num_pi_trajs > 0:
                _z = z.expand(self.cfg.num_pi_trajs)
                for t in range(self.cfg.plan_horizon - 1):
                    pi_actions[t] = self.agent.pi(
                        _z[self.cfg.discrete_encoding_type],
                        eval_mode=not self.cfg.use_prior_policy_noise,
                    )
                    _z = self.model.trans(
                        _z["codes"],
                        pi_actions[t],
                        ce_selection=self.cfg.plan_ce_selection,
                    )
                pi_actions[-1] = self.agent.pi(
                    _z[self.cfg.discrete_encoding_type],
                    eval_mode=not self.cfg.use_prior_policy_noise,
                )

            # Initialize state and parameters
            z = z.expand(self.cfg.num_samples)
            # mean = torch.zeros(self.cfg.plan_horizon, self.act_dim, device=self.device)
            std = self.cfg.max_std * torch.ones(
                self.cfg.plan_horizon, self.act_dim, device=self.device
            )
            if not t0:
                mean[:-1] = prev_mean[1:]
            if self.cfg.num_pi_trajs > 0:
                actions[:, : self.cfg.num_pi_trajs] = pi_actions

            # Iterate MPPI
            for _ in range(self.cfg.iterations):
                # Sample actions
                actions[:, self.cfg.num_pi_trajs :] = (
                    mean.unsqueeze(1)
                    + std.unsqueeze(1)
                    * torch.randn(
                        self.cfg.plan_horizon,
                        self.cfg.num_samples - self.cfg.num_pi_trajs,
                        self.act_dim,
                        device=std.device,
                    )
                ).clamp(-1, 1)
                # breakpoint()

                # Compute elite actions
                value = self._single_estimate_value(z, actions).nan_to_num_(0)
                elite_idxs = torch.topk(
                    value.squeeze(1), self.cfg.num_elites, dim=0
                ).indices
                elite_value, elite_actions = value[elite_idxs], actions[:, elite_idxs]

                # Update parameters
                max_value = elite_value.max(0)[0]
                score = torch.exp(self.cfg.temperature * (elite_value - max_value))
                score /= score.sum(0)
                mean = torch.sum(score.unsqueeze(0) * elite_actions, dim=1) / (
                    score.sum(0) + 1e-9
                )
                std = torch.sqrt(
                    torch.sum(
                        score.unsqueeze(0) * (elite_actions - mean.unsqueeze(1)) ** 2,
                        dim=1,
                    )
                    / (score.sum(0) + 1e-9)
                )
                if self.cfg.add_exploration_noise_inside_mppi == "instead":
                    std = exploration_std
                elif self.cfg.add_exploration_noise_inside_mppi == "max":
                    std = torch.max(std, exploration_std)
                std.clamp_(self.cfg.min_std, self.cfg.max_std)

            act_dist = torch.distributions.Categorical(score[:, 0])
            act_idx = act_dist.sample()
            # actions = elite_actions[:, act_idx]
            actions = torch.index_select(elite_actions, 1, act_idx)[:, 0, :]
            a, std = actions[0], std[0]
            if not eval_mode:
                if self.cfg.add_exploration_noise_after_mppi == "instead":
                    std = self.agent._pi.action_scale * self.agent.exploration_noise
                elif self.cfg.add_exploration_noise_after_mppi == "max":
                    std = torch.max(
                        std, self.agent._pi.action_scale * self.agent.exploration_noise
                    )
                a += std * torch.randn(self.act_dim, device=std.device)
            return a, mean, std

        if hasattr(self, "_prev_mean") and not t0:
            prev_mean = self._prev_mean
        else:
            prev_mean = torch.empty(
                batch_size,
                self.cfg.plan_horizon,
                self.act_dim,
                device=self.device,
            )
        # single_mppi(z_td[0], actions[0], pi_actions[0], mean[0], prev_mean[0])
        a, new_prev_mean, std = torch.vmap(
            single_mppi, in_dims=(0, 0, 0, 0, 0), randomness="different"
        )(z_td, actions, pi_actions, mean, prev_mean)

        if self.cfg.mppi_use_mean:
            a = new_prev_mean[:, 0]

        # if not eval_mode:
        #     if self.cfg.add_exploration_noise_after_mppi == "add":
        #         a += torch.normal(
        #             0, self.agent._pi.action_scale * self.agent.exploration_noise
        #         )
        #     elif self.cfg.add_exploration_noise_after_mppi == "instead":
        #         a = new_prev_mean[:, 0] + torch.normal(
        #             0, self.agent._pi.action_scale * self.agent.exploration_noise
        #         )
        #     elif self.cfg.add_exploration_noise_after_mppi == "max":
        #         std = torch.max(std, self.agent.exploration_noise)
        #         a = new_prev_mean[:, 0] + torch.normal(
        #             0, self.agent._pi.action_scale * std
        #         )

        self._prev_mean = new_prev_mean
        a.clamp_(self.act_spec_low, self.act_spec_high)
        return a, std

    @torch.no_grad()
    def _single_estimate_value(self, z, actions):
        """Estimate value of a trajectory starting at latent state z and executing given actions."""
        G, discount = 0, 1
        for t in range(self.cfg.plan_horizon):
            reward = self.model.reward(z[self.cfg.discrete_encoding_type], actions[t])
            # reward = math.two_hot_inv(self.model.reward(z, actions[t], task), self.cfg)
            z = self.model.trans(
                z["codes"], actions[t], ce_selection=self.cfg.plan_ce_selection
            )
            G += discount * reward
            discount *= self.cfg.rho
        z_pi = z[self.cfg.discrete_encoding_type]
        return G + discount * self.agent.Q(z_pi, self.agent.pi(z_pi), return_type="avg")

    def metrics(self, batch):
        metrics = self.model.metrics(batch)
        metrics.update({"model": h.calc_mean_opt_moments(self.model_opt)})
        metrics.update({"agent": self.agent.metrics(batch)})

        # TODO add dormant neuron ratio stuff
        # metrics.update(h.calc_dormant_neuron_ratio(batch, agent=self))

        return metrics
