#!/usr/bin/env python3
import copy
import logging
from dataclasses import dataclass, field
from typing import List, Optional

import helper as h
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import utils
import wandb
from helper import soft_update_params
from torchrl.data import BoundedTensorSpec, CompositeSpec
from tqdm import tqdm, trange
from utils import ReplayBuffer, ReplayBufferSamples

from .base import Agent, AgentConfig, EvalMode
from .common import Critic


logger = logging.getLogger(__name__)


@dataclass
class SACConfig(AgentConfig):
    """Config for SAC"""

    """MLP dims for actor/critic/dynamics"""
    mlp_dims: List[int] = field(default_factory=lambda: [512, 512])
    """Learning rate for critic"""
    lr: float = 1e-3
    """Batch size - same for for representation learning and actor/critic"""
    batch_size: int = 256
    """Number of parameter updates per new data, i.e .UTD ratio """
    utd_ratio: int = 1
    """Update actor less frequently than critic"""
    actor_update_freq: int = 2
    """Update critic target slower than every time step"""
    critic_target_update_frequency: int = 2
    """Discount factor"""
    gamma: float = 0.99
    """Target network update rate"""
    tau: float = 0.005
    """Number of critics"""
    num_critics: int = 2
    """Number of critics to sample for min/avg"""
    q_sample_size: int = 2
    """Use N-step returns for Q-learning?"""
    nstep: int = 1  # nstep returns
    """What observation types to use? ["state"] or ["pixels"] or ["state", "pixels"]"""
    obs_types: List[str] = field(default_factory=lambda: ["state"])
    """SAC stuff"""
    """Learning rate for actor"""
    policy_lr: float = 3e-4
    """Entropy regularization coefficient."""
    alpha: float = 0.2
    """Auto tune entropy coefficient"""
    autotune: bool = True
    log_std_min: int = -5
    log_std_max: int = 2
    """OTHER"""
    """Logging frequency"""
    logging_freq: int = 100
    """If True try to compile all NNs"""
    compile: bool = False
    """All NNs will be put on this device"""
    device: str = "cuda"
    name: str = "SAC"


class Actor(nn.Module):
    def __init__(self, cfg: SACConfig, state_dim: int, act_spec: BoundedTensorSpec):
        super().__init__()
        self.cfg = cfg
        self.act_dim = np.array(act_spec.shape).prod().item()
        self.action_scale = (act_spec.high - act_spec.low).to(cfg.device) / 2.0
        self.action_bias = (act_spec.high + act_spec.low).to(cfg.device) / 2.0
        self.mlp = h.mlp(state_dim, self.cfg.mlp_dims, 2 * self.act_dim)

    def forward(self, obs):
        mean, log_std = self.mlp(obs).chunk(2, dim=-1)
        log_std = torch.tanh(log_std) + 1
        log_std *= (
            0.5 * (self.cfg.log_std_max - self.cfg.log_std_min) + self.cfg.log_std_min
        )
        std = log_std.exp()
        normal = torch.distributions.Normal(mean, std)
        x_t = normal.rsample()
        y_t = torch.tanh(x_t)
        a = y_t * self.action_scale + self.action_bias
        log_prob = normal.log_prob(x_t)

        # Enforcing Action Bound
        log_prob -= torch.log(self.action_scale * (1 - y_t.pow(2)) + 1e-6)
        log_prob = log_prob.sum(1, keepdim=True)
        mean = torch.tanh(mean) * self.action_scale + self.action_bias
        return {"a": a, "log_prob": log_prob, "mean": mean}


class SAC(Agent):
    def __init__(
        self,
        cfg: SACConfig,
        obs_spec: CompositeSpec,
        act_spec: BoundedTensorSpec,
        state_dim: Optional[int] = None,  # Override observation dim (needed for latent)
    ):
        super().__init__(cfg=cfg, obs_spec=obs_spec, act_spec=act_spec)

        if "pixels" in cfg.obs_types:
            raise NotImplementedError

        ##### Calculate dimensions for MLPs #####
        act_dim = np.array(act_spec.shape).prod().item()
        if "state" in cfg.obs_types:
            if state_dim is not None:
                obs_dim = state_dim
            else:
                obs_dim = np.array(obs_spec["state"].shape).prod().item()
        else:
            raise NotImplementedError("Need to use state observations")

        self.cfg = cfg

        self._pi = Actor(cfg, state_dim=obs_dim, act_spec=act_spec).to(cfg.device)

        ##### Init critics and their target networks #####
        Q = Critic(cfg, state_dim=obs_dim, act_dim=act_dim).to(cfg.device)
        self.Q = torch.compile(Q, mode="default") if cfg.compile else Q
        Q_tar = copy.deepcopy(self.Q).requires_grad_(False)
        self.Q_tar = torch.compile(Q_tar, mode="default") if cfg.compile else Q_tar

        ##### Optimizers #####
        self.pi_opt = torch.optim.Adam(self._pi.parameters(), lr=cfg.policy_lr)
        self.q_opt = torch.optim.Adam(self.Q.parameters(), lr=cfg.lr)

        # Automatic entropy tuning
        if cfg.autotune:
            self.target_entropy = -torch.prod(
                torch.Tensor(act_spec.shape).to(cfg.device)
            ).item()
            self.log_alpha = torch.zeros(1, requires_grad=True, device=cfg.device)
            self.alpha = self.log_alpha.exp().item()
            self.a_opt = torch.optim.Adam([self.log_alpha], lr=cfg.lr)
        else:
            self.alpha = cfg.alpha

        # Counters for number of param updates
        self.critic_update_counter = 0
        self.pi_update_counter = 0

    def update(self, replay_buffer: ReplayBuffer, num_new_transitions: int) -> dict:
        """Update representation and SAC at same time"""
        n = int(num_new_transitions * self.cfg.utd_ratio)
        info = {}

        for i in trange(n, desc="Training SAC:", miniters=self.cfg.logging_freq):
            batch = replay_buffer.sample()

            info.update(self.update_step(batch))

            if i % self.cfg.logging_freq == 0:
                if wandb.run is not None:
                    wandb.log(info)

        self.post_update()
        return info

    def update_step(self, batch: ReplayBufferSamples) -> dict:
        """One SAC update"""

        ##### Make nstep returns #####
        nstep_batch = utils.to_nstep(batch, nstep=self.cfg.nstep, gamma=self.cfg.gamma)

        ##### Update critic #####
        info = self.critic_update_step(batch=nstep_batch)

        ##### Update actor less frequently than critic #####
        if self.critic_update_counter % self.cfg.actor_update_freq == 0:
            for _ in range(self.cfg.actor_update_freq):
                info.update(self.pi_update_step(batch=nstep_batch))

        return info

    def critic_update_step(self, batch: ReplayBufferSamples):
        self.critic_update_counter += 1
        self.Q.train()

        # Check batch shapes
        assert batch.rewards.ndim == 1
        assert batch.rewards.shape[0] == batch.observations.shape[0]

        # Make Q target
        with torch.no_grad():
            obs = batch.observations["state"]
            next_obs = batch.next_observations["state"]
            a = batch.actions
            a_next = self._pi(next_obs)

            min_q_next_tar = self.Q_tar(next_obs, a=a_next["a"], return_type="min")[
                ..., 0
            ]
            min_q_next_tar -= self.alpha * a_next["log_prob"][:, 0]

            assert min_q_next_tar.shape == batch.rewards.shape
            next_q_value = (
                batch.rewards
                + (1 - batch.terminateds) * batch.next_state_gammas * min_q_next_tar
            )

        q_values = self.Q(obs, a=a, return_type="all")[..., 0]
        next_q_value = next_q_value.broadcast_to(q_values.shape)
        q_loss = F.mse_loss(q_values, next_q_value)

        ##### Optimize critic #####
        self.q_opt.zero_grad(set_to_none=True)
        q_loss.backward()
        self.q_opt.step()

        ##### Update the target network #####
        if self.critic_update_counter % self.cfg.critic_target_update_frequency == 0:
            soft_update_params(self.Q, self.Q_tar, tau=self.cfg.tau)

        info = {
            "q_loss": q_loss.item(),
            "critic_update_counter": self.critic_update_counter,
        }
        for i in range(self.cfg.num_critics):
            info.update({f"q{i+1}_values": q_values[i].mean().item()})

        self.Q.eval()
        return info

    def pi_update_step(self, batch: ReplayBufferSamples):
        self.pi_update_counter += 1
        self._pi.train()

        obs = batch.observations["state"]
        a = self._pi(obs)
        if self.discrete_acts:
            raise NotImplementedError("Closed-form expectation for discrete acts")
            pi_loss = (
                a["log_prob"].exp()
                * (
                    self.alpha * a["log_prob"]
                    - self.Q(obs, a=a["a"], return_type="min")
                )
            ).mean()
        else:
            pi_loss = (
                self.alpha * a["log_prob"] - self.Q(obs, a=a["a"], return_type="min")
            ).mean()

        ##### Optimize actor #####
        self.pi_opt.zero_grad(set_to_none=True)
        pi_loss.backward()
        self.pi_opt.step()
        self._pi.eval()

        if self.cfg.autotune:
            with torch.no_grad():
                a = self._pi(obs)
            alpha_loss = (
                -self.log_alpha.exp() * (a["log_prob"] + self.target_entropy).detach()
            ).mean()

            self.a_opt.zero_grad()
            alpha_loss.backward()
            self.a_opt.step()
            self.alpha = self.log_alpha.exp().item()

        info = {
            "actor_loss": pi_loss.item(),
            "alpha": self.alpha,
            "actor_update_counter": self.pi_update_counter,
        }
        if self.cfg.autotune:
            info.update({"alpha_loss": alpha_loss.item()})
        return info

    @torch.no_grad()
    def select_action(self, obs, t0: bool = False, eval_mode: EvalMode = False):
        is_flat_obs = False
        if obs.batch_size == torch.Size([]):
            obs = obs.view(1)
            is_flat_obs = True

        a = self.pi(obs["state"], eval_mode=eval_mode)

        a = a[0] if is_flat_obs else a
        return a

    def pi(self, obs, tar: bool = False, eval_mode: bool = False):
        a = self._pi(obs)
        if eval_mode:
            a = a["mean"]
        else:
            a = a["a"]
        # a = a.clamp(self.act_spec_low, self.act_spec_high)
        if torch.isnan(a).any() or torch.isinf(a).any():
            breakpoint()
        return a

    def post_update(self):
        pass

    def metrics(self, batch):
        metrics = {}
        metrics.update({"Q": h.calc_mean_opt_moments(self.q_opt)})
        metrics.update({"pi": h.calc_mean_opt_moments(self.pi_opt)})

        # TODO add dormant neuron ratio stuff
        # metrics.update(h.calc_dormant_neuron_ratio(batch, agent=self))

        return metrics
