#!/usr/bin/env python3
import copy
from functools import partial
import logging
from dataclasses import dataclass, field
from typing import List, Optional

import helper as h
import numpy as np
import torch
import torch.nn.functional as F
import utils
import wandb
from agents.common import Actor, Critic
from helper import soft_update_params
from torchrl.data import BoundedTensorSpec, CompositeSpec
from tqdm import tqdm, trange
from utils import ReplayBuffer, ReplayBufferSamples

from .base import Agent, AgentConfig, EvalMode


logger = logging.getLogger(__name__)


@dataclass
class TD3Config(AgentConfig):
    """Config for TD3"""

    """MLP dims for actor/critic/dynamics"""
    mlp_dims: List[int] = field(default_factory=lambda: [512, 512])
    """Learning rate for actor/critic"""
    lr: float = 3e-4
    """Batch size - same for for representation learning and actor/critic"""
    batch_size: int = 256
    """Number of parameter updates per new data, i.e .UTD ratio """
    utd_ratio: int = 1
    """Update actor less frequently than critic"""
    actor_update_freq: int = 2
    """Discount factor"""
    gamma: float = 0.99
    """Target network update rate"""
    tau: float = 0.005
    """Number of critics"""
    num_critics: int = 2
    """Number of critics to sample for min/avg"""
    q_sample_size: int = 2
    """Use N-step returns for Q-learning?"""
    nstep: Optional[int] = None  # nstep returns
    """What observation types to use? ["state"] or ["pixels"] or ["state", "pixels"]"""
    obs_types: List[str] = field(default_factory=lambda: ["state"])
    """EXPLORATION NOISE SCHEDULE"""
    """Initial variance"""
    exploration_noise_start: float = 1.0
    """Final variance"""
    exploration_noise_end: float = 0.1
    """Number of episodes do decay noise"""
    exploration_noise_num_steps: int = 50
    """POLICY SMOOTHING"""
    """Variance"""
    policy_noise: float = 0.2
    """Clip the noise"""
    noise_clip: float = 0.3
    """Clips the gradient norm of the critic"""
    grad_clip_norm: Optional[int] = 20
    """OTHER"""
    """Logging frequency"""
    logging_freq: int = 100
    """If True try to compile all NNs"""
    compile: bool = False
    """All NNs will be put on this device"""
    device: str = "cuda"
    name: str = "TD3"


class TD3(Agent):
    def __init__(
        self,
        cfg: TD3Config,
        obs_spec: CompositeSpec,
        act_spec: BoundedTensorSpec,
        state_dim: Optional[int] = None,  # Override observation dim (needed for latent)
    ):
        super().__init__(cfg=cfg, obs_spec=obs_spec, act_spec=act_spec)

        if "pixels" in cfg.obs_types:
            raise NotImplementedError

        ##### Calculate dimensions for MLPs #####
        act_dim = np.array(act_spec.shape).prod().item()
        if "state" in cfg.obs_types:
            if state_dim is not None:
                obs_dim = state_dim
            else:
                obs_dim = np.array(obs_spec["state"].shape).prod().item()
        else:
            raise NotImplementedError("Need to use state observations")

        self.cfg = cfg

        self._pi = Actor(
            cfg,
            state_dim=obs_dim,
            act_dim=act_dim,
            action_scale=(act_spec.high - act_spec.low).to(cfg.device) / 2.0,
            action_bias=(act_spec.high + act_spec.low).to(cfg.device) / 2.0,
            act_low=act_spec.low,
            act_high=act_spec.high,
        ).to(cfg.device)
        self._pi = torch.compile(self._pi, mode="default") if cfg.compile else self._pi
        pi_tar = copy.deepcopy(self._pi).requires_grad_(False)
        self._pi_tar = torch.compile(pi_tar, mode="default") if cfg.compile else pi_tar

        ##### Init critics and their target networks #####
        Q = Critic(cfg, state_dim=obs_dim, act_dim=act_dim).to(cfg.device)
        self.Q = torch.compile(Q, mode="default") if cfg.compile else Q
        Q_tar = copy.deepcopy(self.Q).requires_grad_(False)
        self.Q_tar = torch.compile(Q_tar, mode="default") if cfg.compile else Q_tar

        ##### Optimizers #####
        self.pi_opt = torch.optim.Adam(self._pi.parameters(), lr=cfg.lr)
        self.q_opt = torch.optim.Adam(self.Q.parameters(), lr=cfg.lr)

        ##### Exploration noise schedule #####
        self._exploration_noise_schedule = h.LinearSchedule(
            start=cfg.exploration_noise_start,
            end=cfg.exploration_noise_end,
            num_steps=cfg.exploration_noise_num_steps,
            device=cfg.device,
        )

        # Counters for number of param updates
        self.critic_update_counter = 0
        self.pi_update_counter = 0

    def update(self, replay_buffer: ReplayBuffer, num_new_transitions: int) -> dict:
        """Update representation and TD3 at same time"""
        n = int(num_new_transitions * self.cfg.utd_ratio)
        info = {}

        for i in trange(
            n, desc=f"Training {self.name}:", miniters=self.cfg.logging_freq
        ):
            batch = replay_buffer.sample()

            info.update(self.update_step(batch))

            if i % self.cfg.logging_freq == 0:
                if wandb.run is not None:
                    wandb.log(info)

        self.post_update()
        return info

    # @partial(torch.compile)(mode="reduce-overhead", fullgraph=True)
    def update_step(self, batch: ReplayBufferSamples) -> dict:
        """One TD3 update"""

        ##### Make nstep returns (or flatten) #####
        if self.cfg.nstep is None and batch.rewards.ndim == 2:
            # Flatten [H, B] into [H*B]
            batch = utils.flatten_batch(batch)
        else:
            batch = utils.to_nstep(batch, nstep=self.cfg.nstep, gamma=self.cfg.gamma)

        ##### Update critic #####
        info = self.critic_update_step(batch=batch)

        ##### Update actor less frequently than critic #####
        if self.critic_update_counter % self.cfg.actor_update_freq == 0:
            info.update(self.pi_update_step(batch=batch))

        info.update({"exploration_noise": self.exploration_noise})
        return info

    # @partial(torch.compile)(mode="reduce-overhead", fullgraph=True)
    def critic_update_step(self, batch: ReplayBufferSamples):
        self.critic_update_counter += 1
        self.Q.train()

        # Check batch shapes
        assert batch.rewards.ndim == 1
        assert batch.rewards.shape[0] == batch.observations.shape[0]

        # Make Q target
        with torch.no_grad():
            obs = batch.observations["state"]
            next_obs = batch.next_observations["state"]
            a = batch.actions
            a_next = self.pi(next_obs, tar=True, eval_mode=True, smooth=True)

            min_q_next_tar = self.Q_tar(next_obs, a=a_next, return_type="min")[..., 0]

            assert min_q_next_tar.shape == batch.rewards.shape
            next_q_value = (
                batch.rewards
                + (1 - batch.terminateds) * batch.next_state_gammas * min_q_next_tar
            )

        q_values = self.Q(obs, a=a, return_type="all")[..., 0]
        next_q_value = next_q_value.broadcast_to(q_values.shape)
        q_loss = F.mse_loss(q_values, next_q_value)
        info = {
            "q_loss": q_loss.item(),
            "q_mean": q_values.mean().item(),
            "q_min": q_values.min().item(),
            "q_max": q_values.max().item(),
            "q_std": q_values.std().item(),
            "q_targ_mean": next_q_value.mean().item(),
            "q_targ_min": next_q_value.min().item(),
            "q_targ_max": next_q_value.max().item(),
            "q_targ_std": next_q_value.std().item(),
            "critic_update_counter": self.critic_update_counter,
        }

        ##### Optimize critic #####
        self.q_opt.zero_grad(set_to_none=True)
        q_loss.backward()

        if self.cfg.grad_clip_norm is not None:
            q_params = list(self.Q.parameters())
            grad_norm = torch.nn.utils.clip_grad_norm_(
                q_params, self.cfg.grad_clip_norm, error_if_nonfinite=False
            )
            info.update({"q_grad_norm": float(grad_norm)})

        self.q_opt.step()

        ##### Update the target network #####
        soft_update_params(self.Q, self.Q_tar, tau=self.cfg.tau)

        for i in range(self.cfg.num_critics):
            info.update({f"q{i+1}_values": q_values[i].mean().item()})

        self.Q.eval()
        return info

    # @partial(torch.compile)(mode="reduce-overhead", fullgraph=True)
    def pi_update_step(self, batch: ReplayBufferSamples):
        self.pi_update_counter += 1
        self._pi.train()

        obs = batch.observations["state"]
        pi_loss = -self.Q(obs, a=self._pi(obs), return_type="avg").mean()
        info = {
            "actor_loss": pi_loss.item(),
            "actor_update_counter": self.pi_update_counter,
        }

        ##### Optimize actor #####
        self.pi_opt.zero_grad(set_to_none=True)
        pi_loss.backward()

        if self.cfg.grad_clip_norm is not None:
            pi_params = list(self._pi.parameters())
            grad_norm = torch.nn.utils.clip_grad_norm_(
                pi_params, self.cfg.grad_clip_norm, error_if_nonfinite=False
            )
            info.update({"pi_grad_norm": float(grad_norm)})

        self.pi_opt.step()

        ##### Update the target network #####
        soft_update_params(self._pi, self._pi_tar, tau=self.cfg.tau)

        self._pi.eval()
        return info

    # @torch.compile
    @torch.no_grad()
    def select_action(self, obs, t0: bool = False, eval_mode: EvalMode = False):
        is_flat_obs = False
        if obs.batch_size == torch.Size([]):
            obs = obs.view(1)
            is_flat_obs = True

        a = self.pi(obs["state"], tar=False, eval_mode=eval_mode)

        a = a[0] if is_flat_obs else a
        return a

    def pi(self, obs, tar: bool = False, eval_mode: bool = False, smooth: bool = False):
        a = self._pi_tar(obs) if tar else self._pi(obs)
        if not eval_mode:
            a += torch.normal(0, self._pi.action_scale * self.exploration_noise)
        if smooth:
            clipped_noise = (
                torch.randn_like(a, device=self.cfg.device) * self.cfg.policy_noise
            ).clamp(-self.cfg.noise_clip, self.cfg.noise_clip) * self._pi.action_scale
            a += clipped_noise
        a = a.clamp(self.act_spec_low, self.act_spec_high)
        return a

    def post_update(self):
        self._exploration_noise_schedule.step()

    @property
    def exploration_noise(self):
        return self._exploration_noise_schedule()

    def metrics(self, batch):
        metrics = {}
        metrics.update({"Q": h.calc_mean_opt_moments(self.q_opt)})
        metrics.update({"pi": h.calc_mean_opt_moments(self.pi_opt)})

        # TODO add dormant neuron ratio stuff
        # metrics.update(h.calc_dormant_neuron_ratio(batch, agent=self))

        return metrics
