#!/usr/bin/env python3
from typing import Optional

from .dmcontrol import make_env as dmcontrol_make_env
import gymnasium as gym
from dm_control import suite
from metaworld.envs import ALL_V2_ENVIRONMENTS_GOAL_OBSERVABLE
from torchrl.envs import (
    DMControlEnv,
    GymEnv,
    ParallelEnv,
    SerialEnv,
    StepCounter,
    TransformedEnv,
)
from torchrl.envs.transforms import (
    CatFrames,
    CatTensors,
    Compose,
    DoubleToFloat,
    RenameTransform,
    Resize,
    RewardSum,
    ToTensorImage,
    TransformedEnv,
)

# from torchrl.record import PixelRenderTransform, VideoRecorder
from torchrl.record.loggers.csv import CSVLogger

from .metaworld import make_env as metaworld_make_env


# import shimmy


def make_env(cfg, num_envs: int = 1, record_video: bool = True):
    # cfg.from_pixels = True
    # cfg.render_size = 64
    if num_envs == 1:
        env = SerialEnv(
            num_envs,
            make_env_fn,
            # create_env_kwargs={"record_video": record_video},
            create_env_kwargs={"cfg": cfg},
            device=cfg.device,  # puts policy on device
        )
    else:
        # parallel_env_fn = ParallelEnv
        parallel_env_fn = ParallelEnv if not cfg.debug else SerialEnv
        # eval_kwargs = [{"record_video": False}] * cfg.num_eval_episodes
        # eval_kwargs[0] = {"record_video": cfg.capture_eval_video}
        env = parallel_env_fn(
            num_envs,
            make_env_fn,
            # create_env_kwargs=eval_kwargs,
            create_env_kwargs={"cfg": cfg},
            serial_for_single=True,
            device=cfg.device,  # puts policy on device
        )
    env.set_seed(cfg.seed)

    # breakpoint()
    # if record_video:
    #     # env = env.append_transform(PixelRenderTransform())
    #     logger = CSVLogger("dummy", video_format="mp4")
    #     env = env.append_transform(
    #         PixelRenderTransform(mode="rgb_array", preproc=lambda x: x.copy())
    #     )
    #     env = env.append_transform(VideoRecorder(logger=logger, tag="pixels_record"))
    # TransformedEnv(
    #     env,
    #     Compose(
    #         DoubleToFloat(),
    #         StepCounter(),
    #         RewardSum(),
    #         PixelRenderTransform(mode="rgb_array", preproc=lambda x: x.copy()),
    #     ),
    # )
    # env.start()
    return env


def make_env_fn(
    cfg,
    # record_video: bool = False,
    # env_name: str,
    # num_envs: int = 1,
    # env_id: Optional[str] = None,
    # task_name: Optional[str] = None,
    # seed: int = 42,
    # from_pixels: bool = True,
    # frame_skip: int = 2,
    # pixels_only: bool = False,
    # render_size: int = 64,
    # num_frames_to_stack: int = 1,
    # logger=None,
    # record_video: bool = False,
    # device: str = "cpu",
    # max_episode_steps: int = 500,
):
    record_video = False
    logger = None
    if not cfg.from_pixels:
        cfg.pixels_only = False

    if cfg.env_name in gym.envs.registry.keys():
        env = GymEnv(
            env_name=cfg.env_name,
            from_pixels=cfg.from_pixels,
            frame_skip=cfg.action_repeat,
            pixels_only=cfg.pixels_only,
            # terminate_when_unhealthy=False,
            device=cfg.env_device,
        )
    elif (cfg.env_name, cfg.task_name) in suite.ALL_TASKS or cfg.env_name == "cup":
        env = dmcontrol_make_env(
            env_name=cfg.env_name,
            task_name=cfg.task_name,
            from_pixels=cfg.from_pixels or record_video,
            frame_skip=cfg.action_repeat,
            pixels_only=cfg.pixels_only,
            device=cfg.env_device,
        )
    elif (
        cfg.env_name.split("-", 1)[-1] + "-v2-goal-observable"
        in ALL_V2_ENVIRONMENTS_GOAL_OBSERVABLE
    ):
        env = metaworld_make_env(
            env_name=cfg.env_name.split("-", 1)[-1] + "-v2-goal-observable",
            from_pixels=cfg.from_pixels or record_video,
            seed=cfg.seed,
            frame_skip=cfg.action_repeat,
            pixels_only=cfg.pixels_only,
            device=cfg.env_device,
            max_episode_steps=cfg.max_episode_steps,
        )

    if not cfg.pixels_only:
        env = TransformedEnv(
            env,
            Compose(
                RenameTransform(in_keys=["observation"], out_keys=["state"]),
                RenameTransform(in_keys=["state"], out_keys=[("observation", "state")]),
            ),
        )
    env = TransformedEnv(
        env,
        Compose(DoubleToFloat(), StepCounter(), RewardSum()),
    )

    if cfg.from_pixels:
        env = TransformedEnv(
            env,
            Compose(
                ToTensorImage(in_keys="pixels"),
                Resize(cfg.render_size, cfg.render_size),
                # RenameTransform(in_keys="pixels", out_keys=("observation", "pixels")),
                RenameTransform(
                    in_keys=["pixels"], out_keys=[("observation", "pixels")]
                ),
                CatFrames(
                    N=cfg.num_frames_to_stack, dim=-3, in_keys=("observation", "pixels")
                ),
            ),
        )
    # video_rec_in_keys = ("observation", "pixels")
    # else:
    # video_rec_in_keys = "pixels"
    # if record_video:
    #     if logger is None:
    #         logger = CSVLogger(exp_name="", log_dir="./logs", video_format="mp4")
    #     env = TransformedEnv(
    #         env,
    #         VideoRecorder(logger=logger, tag="run_video", in_keys=video_rec_in_keys),
    #     )

    # env.set_seed(seed)
    return env
