#!/usr/bin/env python3
import os
import sys
from dataclasses import dataclass, field
from functools import partial
from typing import Optional

import hydra
import omegaconf
from agents.base import AgentConfig
from agents.fsq_mpc import FSQMPCConfig
from hydra.core.config_store import ConfigStore


# @dataclass
# class AgentConfig:
#     _target_: str = "agents.TD3"
#     device: str = "cuda"


# @dataclass
# class TrainConfig:
#     agent: dict
#     env_name: str = "cartpole"
#     task_name: str = "swingup"
#     env_id: Optional[str] = None  # Used for old gym envs

#     # This should work with hydra.utils.instantiate(cfg.agent,)
#     # agent: FSQMPCConfig = field(default_factory=FSQMPCConfig)
#     # agent: AgentConfig = field(default_factory=AgentConfig)

#     # Observation stuff
#     from_pixels: bool = False
#     pixels_only: bool = False
#     num_frames_to_stack: int = 3  # only used for pixel observations

#     # Experiment
#     max_episode_steps: int = 500  # Max episode length (1000 steps as action_repeat=2)
#     num_episodes: int = 500  # Number of training episodes
#     # random_episodes: int = 10  # Number of random episodes at start
#     random_env_steps: int = 5_000  # Number of random env steps at start
#     action_repeat: int = 2
#     buffer_size: int = 10_000_000
#     prefetch: int = 5
#     seed: int = 42
#     checkpoint: Optional[str] = None  # /file/path/to/checkpoint
#     device: str = "cuda"  # "cpu" or "cuda" etc
#     env_device: str = "cpu"  # DMControl/MetaWorld on cpu but maniskill/isaac on cuda
#     debug: bool = False

#     render_size: int = 64

#     # Evaluation
#     eval_every_episodes: int = 10
#     num_eval_episodes: int = 10
#     capture_eval_video: bool = True  # Fails on AMD GPU so set to False
#     capture_train_video: bool = False
#     log_dormant_neuron_ratio: bool = False

#     # W&B config
#     use_wandb: bool = False
#     wandb_project_name: str = "DCWM"
#     run_name: str = f"FSQ-MPC"


# cs = ConfigStore.instance()
# cs.store(name="base_train_config", node=TrainConfig)
# cs.store(name="base_fsqmpc", group="agent", node=FSQMPCConfig)


@hydra.main(version_base="1.3", config_path="./cfgs", config_name="train")
def train(cfg):
    # def train(cfg: TrainConfig):
    import logging
    import pdb

    if cfg.debug:
        pdb.set_trace()
    import random
    import time

    # This is needed to render videos on GPU
    # if torch.cuda.is_available() and (cfg.device == "cuda"):
    #     os.environ["MUJOCO_GL"] = "osmesa"
    #     os.environ["PYOPENGL_PLATFORM"] = "osmesa"
    import agents
    import numpy as np
    import torch
    from envs import make_env
    from hydra.utils import get_original_cwd
    from omegaconf import DictConfig
    from tensordict.nn import TensorDictModule
    from termcolor import colored
    from torchrl.data.tensor_specs import BoundedTensorSpec
    from torchrl.record.loggers.wandb import WandbLogger
    from utils import ReplayBuffer

    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger(__name__)

    ###### Fix seed for reproducibility ######
    random.seed(cfg.seed)
    np.random.seed(cfg.seed)
    torch.manual_seed(cfg.seed)
    torch.backends.cudnn.deterministic = True
    # torch.backends.cudnn.benchmark = False

    cfg.device = (
        "cuda" if torch.cuda.is_available() and (cfg.device == "cuda") else "cpu"
    )
    cfg.agent.device = cfg.device
    if cfg.agent.r_max is not None:
        cfg.agent.r_max *= cfg.action_repeat

    ###### Initialise W&B ######
    writer = WandbLogger(
        exp_name=cfg.run_name,
        offline=not cfg.use_wandb,
        project=cfg.wandb_project_name,
        # log_dir="./logs",
        group=f"{cfg.env_name}-{cfg.task_name}",
        tags=[f"{cfg.env_name}-{cfg.task_name}", f"seed={str(cfg.seed)}"],
        #     monitor_gym=cfg.monitor_gym,
        save_code=True,
        # dir=os.path.join(get_original_cwd(), "output"),
        # dir="./logs",
    )
    writer.log_hparams(cfg)

    ###### Setup vectorized environment for training/evaluation/video recording ######
    env = make_env(cfg, num_envs=1, record_video=cfg.capture_train_video)
    eval_env = make_env(
        cfg, num_envs=cfg.num_eval_episodes, record_video=cfg.capture_eval_video
    )
    assert isinstance(
        env.action_spec, BoundedTensorSpec
    ), "only continuous act space supported"

    ###### Prepare replay buffer ######
    if cfg.agent.nstep is not None:
        nstep = max(cfg.agent.get("nstep", 1), cfg.agent.get("horizon", 1))
    else:
        nstep = cfg.agent.get("horizon", 1)
    rb = ReplayBuffer(
        buffer_size=cfg.buffer_size,
        batch_size=cfg.agent.batch_size,
        nstep=nstep,
        gamma=cfg.agent.gamma,
        prefetch=cfg.prefetch,
        pin_memory=True,  # will be set to False if device=="cpu"
        device=cfg.device,
    )

    ###### Init agent ######
    agent = hydra.utils.instantiate(
        DictConfig({"_target_": cfg.agent._target_, "cfg": cfg.agent}),
        _partial_=True,
        _recursive_=False,
    )(obs_spec=env.observation_spec["observation"][0], act_spec=env.action_spec[0])

    # Load state dict into this agent from filepath (or dictionary)
    if cfg.checkpoint is not None:
        state_dict = torch.load(os.path.join(get_original_cwd(), cfg.checkpoint))
        agent.load_state_dict(state_dict["model"])
        logger.info(f"Loaded checkpoint from {cfg.checkpoint}")

    # Print information about run
    task = cfg.env_name if cfg.task_name == "" else cfg.env_name + "-" + cfg.task_name
    steps = (cfg.num_episodes * cfg.max_episode_steps) / 1e6
    total_params = int(agent.total_params / 1e6)
    writer.log_hparams({"total_params": agent.total_params})
    print(colored("Task:", "yellow", attrs=["bold"]), task)
    print(colored("Number of episodes:", "yellow", attrs=["bold"]), cfg.num_episodes)
    print(colored("Max number of env. steps:", "yellow", attrs=["bold"]), steps, "M")
    print(colored("Action repeat:", "green", attrs=["bold"]), cfg.action_repeat)
    print(colored("Device:", "green", attrs=["bold"]), cfg.device)
    print(colored("Learnable parameters:", "green", attrs=["bold"]), f"{total_params}M")
    print(colored("Architecture:", "green", attrs=["bold"]), agent)

    def build_policy_module(eval_mode):
        return TensorDictModule(
            lambda obs, step_count: agent.select_action(
                obs, t0=step_count[0] == 0, eval_mode=eval_mode
            ),
            in_keys=["observation", "step_count"],
            out_keys=["action"],
        )

    policy_module = build_policy_module(eval_mode=False)
    eval_policy_module = build_policy_module(eval_mode=True)

    def evaluate():
        """Calculate avg. episodic return (optionally avg. success)"""
        eval_metrics = {}
        with torch.no_grad():
            eval_start_time = time.time()
            eval_data = eval_env.rollout(
                max_steps=cfg.max_episode_steps, policy=eval_policy_module
            )
            print(f"Finished rollout in eval_env {time.time()-eval_start_time}")
            episodic_return = torch.mean(eval_data["next"]["episode_reward"][:, -1, 0])
            success = eval_data["next"].get("success", None)
            if success is not None:
                episodic_success = torch.mean(success.any(-1).to(torch.float))
                eval_metrics.update({"episodic_success": episodic_success})
            if cfg.capture_eval_video:
                video_env.rollout(
                    max_steps=cfg.max_episode_steps,
                    policy=eval_policy_module,
                )
                video_env.transform.dump()

        ##### Eval metrics #####
        eval_metrics.update(
            {
                "episodic_return": episodic_return,
                "elapsed_time": time.time() - start_time,
                "SPS": int(env_step / (time.time() - start_time)),
                "episode_time": time.time() - episode_start_time,
                "env_step": env_step * cfg.action_repeat,
                "step": env_step,
                "episode": episode_idx,
                "train_time": train_time,
                "rollout_time": rollout_time,
            }
        )
        logger.info(
            f"Eval | Return {episode_reward:.2f} | Env step {env_step*cfg.action_repeat} | Episode {episode_idx} | SPS {eval_metrics['SPS']}"
        )

        ##### Log rank of latent and active codebook percent #####
        batch = rb.sample(batch_size=agent.model.cfg.latent_dim)
        eval_metrics.update(agent.metrics(batch))

        # ##### Log dormant neuron ratio #####
        # if cfg.log_dormant_neuron_ratio:
        #     eval_metrics.update(h.calc_dormant_neuron_ratio(batch, agent=agent))

        ##### Log metrics to W&B or csv #####
        writer.log_scalar(name="eval/", value=eval_metrics)

    env_step, start_time, train_time = 0, time.time(), 0
    for episode_idx in range(cfg.num_episodes):
        episode_start_time = time.time()
        ##### Rollout the policy in the environment #####
        with torch.no_grad():
            if env_step < cfg.random_env_steps:
                data = env.rollout(max_steps=cfg.max_episode_steps)[0]
            else:
                data = env.rollout(
                    max_steps=cfg.max_episode_steps, policy=policy_module
                )[0]
        if episode_idx == 0:
            print(colored("First episodes data:", "green", attrs=["bold"]), data)

        ##### Add data to the replay buffer #####
        rb.extend(data)

        ##### Log episode metrics #####
        rollout_time = time.time() - episode_start_time
        num_new_transitions = data["next"]["step_count"][-1].cpu().item()
        env_step += num_new_transitions
        episode_reward = data["next"]["episode_reward"][-1].cpu().item()
        logger.info(
            f"Train | Return {episode_reward:.2f} | Env Step {env_step*cfg.action_repeat} | Episode {episode_idx} | New data {num_new_transitions}"
        )
        rollout_metrics = {
            "episodic_return": episode_reward,
            "episodic_length": num_new_transitions,
            "env_step": env_step * cfg.action_repeat,
        }
        success = data["next"].get("success", None)
        if success is not None:
            episode_success = success.any()
            rollout_metrics.update({"episodic_success": episode_success})

        writer.log_scalar(name="rollout/", value=rollout_metrics)

        ##### Train agent (after collecting some random episodes) #####
        # if (episode_idx > cfg.random_episodes - 1):
        if env_step > cfg.random_env_steps - 1:
            # logger.info(
            #     f"Training agent w/ {num_new_transitions} new data @ step {env_step}..."
            # )
            tol = cfg.max_episode_steps / 2
            if cfg.random_env_steps - tol <= env_step <= cfg.random_env_steps + tol:
                # num_new_transitions = max(1000, 5 * cfg.max_episode_steps)
                num_new_transitions = cfg.random_env_steps
            train_start_time = time.time()
            train_metrics = agent.update(
                replay_buffer=rb, num_new_transitions=num_new_transitions
            )
            train_time = time.time() - train_start_time
            # logger.info("Finished training agent.")

            ##### Log training metrics #####
            writer.log_scalar(name="train/", value=train_metrics)

            ##### Save checkpoint #####
            torch.save({"model": agent.state_dict()}, "./checkpoint")

        ###### Evaluate ######
        if episode_idx % cfg.eval_every_episodes == 0:
            evaluate()

        # Release some GPU memory (if possible)
        torch.cuda.empty_cache()

    env.close()
    eval_env.close()


if __name__ == "__main__":
    train()  # pyright: ignore
